@@ -1,5 +1,5 @@
-/* cabextract 1.4 - a program to extract Microsoft Cabinet files
- * (C) 2000-2011 Stuart Caie <kyzer@4u.net>
+/* cabextract - a program to extract Microsoft Cabinet files
+ * (C) 2000-2018 Stuart Caie <kyzer@cabextract.org.uk>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,103 +17,47 @@
  */
 
 /* cabextract uses libmspack to access cabinet files. libmspack is
- * available from http://www.cabextract.org.uk/libmspack/
+ * available from https://www.cabextract.org.uk/libmspack/
  */
 
 #define _GNU_SOURCE 1
 
 #if HAVE_CONFIG_H
-#include <config.h>
-
-#include <stdio.h> /* everyone has this! */
-
-#if HAVE_SYS_TYPES_H
-# include <sys/types.h>
+# include <config.h>
 #endif
 
-#if HAVE_CTYPE_H
-# include <ctype.h>
-#endif
+#include <sys/types.h>
 
-#if HAVE_ERRNO_H
-# include <errno.h>
+#include <ctype.h>
+#include <dirent.h>
+#include <errno.h>
+#include <fnmatch.h>
+#include <limits.h>
+#include <locale.h>
+#include <stdarg.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#ifdef HAVE_STRINGS_H
+# include <strings.h> /* BSD defines strcasecmp() here */
 #endif
+#include <sys/stat.h>
+#include <time.h>
 
-#if HAVE_FNMATCH_H
-# include <fnmatch.h>
+#if HAVE_ICONV
+# include <iconv.h>
 #endif
-
-#if HAVE_LIMITS_H
-# include <limits.h>
+#if HAVE_TOWLOWER
+# include <wctype.h>
 #endif
-
-#if HAVE_STDARG_H
-# include <stdarg.h>
+#if HAVE_UTIME
+# include <utime.h>
 #endif
-
-#if HAVE_STDLIB_H
-# include <stdlib.h>
-#endif
-
-#if HAVE_STRING_H
-# include <string.h>
-#endif
-
-#if HAVE_STRINGS_H
-# include <strings.h>
-#endif
-
-#if HAVE_SYS_STAT_H
-# include <sys/stat.h>
-#endif
-
-#if TIME_WITH_SYS_TIME
+#if HAVE_UTIMES
 # include <sys/time.h>
-# include <time.h>
-#else
-# if HAVE_SYS_TIME_H
-#  include <sys/time.h>
-# else
-#  include <time.h>
-# endif
-#endif
-
-#if HAVE_UTIME || HAVE_UTIMES
-# if HAVE_UTIME_H
-#  include <utime.h>
-# else
-#  include <sys/utime.h>
-# endif
-#endif
-
-#if HAVE_DIRENT_H
-# include <dirent.h>
-#else
-# define dirent direct
-# if HAVE_SYS_NDIR_H
-#  include <sys/ndir.h>
-# endif
-# if HAVE_SYS_DIR_H
-#  include <sys/dir.h>
-# endif
-# if HAVE_NDIR_H
-#  include <ndir.h>
-# endif
-#endif
-
-#if !STDC_HEADERS
-# if !HAVE_STRCHR
-#  define strchr index
-#  define strrchr rindex
-# endif
-# if !HAVE_STRCASECMP
-#  define strcasecmp strcmpi
-# endif
-# if !HAVE_MEMCPY
-#  define memcpy(d,s,n) bcopy((s),(d),(n))
-# endif
 #endif
 
+/* ensure mkdir(pathname, mode) exists */
 #if HAVE_MKDIR
 # if MKDIR_TAKES_ONE_ARG
 #  define mkdir(a, b) mkdir(a)
@@ -126,24 +70,21 @@
 # endif
 #endif
 
-#ifndef HAVE_MKTIME
-extern time_t mktime(struct tm *tp);
-#endif
-
 #ifndef FNM_CASEFOLD
 # define FNM_CASEFOLD (0)
 #endif
 
 #include "getopt.h"
 
-#endif
-
 #include <mspack.h>
 #include <md5.h>
 
 /* structures and global variables */
 struct option optlist[] = {
   { "directory", 1, NULL, 'd' },
+#if HAVE_ICONV
+  { "encoding",  1, NULL, 'e' },
+#endif
   { "fix",       0, NULL, 'f' },
   { "filter",    1, NULL, 'F' },
   { "help",      0, NULL, 'h' },
@@ -157,6 +98,12 @@ struct option optlist[] = {
   { NULL,        0, NULL, 0   }
 };
 
+#if HAVE_ICONV
+const char *OPTSTRING = "d:e:fF:hlLpqstv";
+#else
+const char *OPTSTRING = "d:fF:hlLpqstv";
+#endif
+
 struct file_mem {
   struct file_mem *next;
   dev_t st_dev;
@@ -166,7 +113,7 @@ struct file_mem {
 
 struct cabextract_args {
   int help, lower, pipe, view, quiet, single, fix, test;
-  char *dir, *filter;
+  char *dir, *filter, *encoding;
 };
 
 /* global variables */
@@ -176,13 +123,16 @@ struct file_mem *cab_args = NULL;
 struct file_mem *cab_exts = NULL;
 struct file_mem *cab_seen = NULL;
 
-mode_t user_umask;
+mode_t user_umask = 0;
 
 struct cabextract_args args = {
   0, 0, 0, 0, 0, 0, 0, 0,
-  NULL, NULL
+  NULL, NULL, NULL
 };
 
+#if HAVE_ICONV
+iconv_t converter = NULL;
+#endif
 
 /** A special filename. Extracting to this filename will send the output
  * to standard output instead of a file on disk. The magic happens in
@@ -196,7 +146,6 @@ const char *STDOUT_FNAME = "stdout";
  * magic happens in cabx_open() when the TEST_FNAME pointer is given as a
  * filename, so treat this like a constant rather than a string. 
  */
-
 const char *TEST_FNAME = "test";
 
 /** A global MD5 context, used when a file is written to TEST_FNAME */
@@ -212,10 +161,17 @@ static void load_spanning_cabinets(struct mscabd_cabinet *basecab,
                                    char *basename);
 static char *find_cabinet_file(char *origcab, char *cabname);
 static int unix_path_seperators(struct mscabd_file *files);
-static char *create_output_name(unsigned char *fname, unsigned char *dir,
+static char *create_output_name(const char *fname, const char *dir,
                                 int lower, int isunix, int unicode);
 static void set_date_and_perm(struct mscabd_file *file, char *filename);
 
+#if HAVE_ICONV
+static void convert_filenames(struct mscabd_file *files);
+#endif
+#if LATIN1_FILENAMES
+static void convert_utf8_to_latin1(char *str);
+#endif
+
 static void memorise_file(struct file_mem **fml, char *name, char *from);
 static int recall_file(struct file_mem *fml, char *name, char **from);
 static void forget_files(struct file_mem **fml);
@@ -247,20 +203,36 @@ static struct mspack_system cabextract_system = {
 int main(int argc, char *argv[]) {
   int i, err;
 
+   /* attempt to set a UTF8-based locale, so that tolower()/towlower()
+    * in create_output_name() lowercase more than just A-Z in ASCII.
+    *
+    * We don't attempt to pick up the system default locale, "",
+    * because it might not be compatible with ASCII/ISO-8859-1/Unicode
+    * character codes and would mess up lowercased filenames
+    */
+   char *locales[] = {
+       "C.UTF-8", /* https://sourceware.org/glibc/wiki/Proposals/C.UTF-8 */
+       "en_US.UTF-8", "en_GB.UTF8", "de_DE.UTF-8", "UTF-8", "UTF8"
+   };
+   for (i = 0; i < (sizeof(locales)/sizeof(*locales)); i++) {
+      if (setlocale(LC_CTYPE, locales[i])) break;
+   }
+
   /* parse options */
-  while ((i = getopt_long(argc, argv, "d:fF:hlLpqstv", optlist, NULL)) != -1) {
+  while ((i = getopt_long(argc, argv, OPTSTRING, optlist, NULL)) != -1) {
     switch (i) {
-    case 'd': args.dir    = optarg; break;
-    case 'f': args.fix    = 1;      break;
-    case 'F': args.filter = optarg; break;
-    case 'h': args.help   = 1;      break;
-    case 'l': args.view   = 1;      break;
-    case 'L': args.lower  = 1;      break;
-    case 'p': args.pipe   = 1;      break;
-    case 'q': args.quiet  = 1;      break;
-    case 's': args.single = 1;      break;
-    case 't': args.test   = 1;      break;
-    case 'v': args.view   = 1;      break;
+    case 'd': args.dir      = optarg; break;
+    case 'e': args.encoding = optarg; break;
+    case 'f': args.fix      = 1;      break;
+    case 'F': args.filter   = optarg; break;
+    case 'h': args.help     = 1;      break;
+    case 'l': args.view     = 1;      break;
+    case 'L': args.lower    = 1;      break;
+    case 'p': args.pipe     = 1;      break;
+    case 'q': args.quiet    = 1;      break;
+    case 's': args.single   = 1;      break;
+    case 't': args.test     = 1;      break;
+    case 'v': args.view     = 1;      break;
     }
   }
 
@@ -278,13 +250,16 @@ int main(int argc, char *argv[]) {
       "  -t   --test        test cabinet integrity\n"
       "  -q   --quiet       only print errors and warnings\n"
       "  -L   --lowercase   make filenames lowercase\n"
-      "  -f   --fix         fix (some) corrupted cabinets\n");
+      "  -f   --fix         salvage as much as possible from corrupted cabinets\n");
     fprintf(stderr,
       "  -p   --pipe        pipe extracted files to stdout\n"
       "  -s   --single      restrict search to cabs on the command line\n"
       "  -F   --filter      extract only files that match the given pattern\n"
+#if HAVE_ICONV
+      "  -e   --encoding    assume non-UTF8 filenames have the given encoding\n"
+#endif
       "  -d   --directory   extract all files to the given directory\n\n"
-      "cabextract %s (C) 2000-2011 Stuart Caie <kyzer@4u.net>\n"
+      "cabextract %s (C) 2000-2018 Stuart Caie <kyzer@cabextract.org.uk>\n"
       "This is free software with ABSOLUTELY NO WARRANTY.\n",
       VERSION);
     return EXIT_FAILURE;
@@ -339,11 +314,25 @@ int main(int argc, char *argv[]) {
   }
 
   /* obtain user's umask */
-  user_umask = umask(0);
-  umask(user_umask);
+#if HAVE_UMASK
+  umask(user_umask = umask(0));
+#endif
 
-  /* turn on/off 'fix MSZIP' mode */
+  /* turn on/off 'fix MSZIP' and 'salvage' mode */
   cabd->set_param(cabd, MSCABD_PARAM_FIXMSZIP, args.fix);
+  cabd->set_param(cabd, MSCABD_PARAM_SALVAGE, args.fix);
+
+#if HAVE_ICONV
+  /* set up converter for given encoding */
+    if (args.encoding) {
+      if ((converter = iconv_open("UTF8", args.encoding)) == (iconv_t) -1) {
+        converter = NULL;
+        fprintf(stderr, "FATAL ERROR: encoding '%s' is not recognised\n",
+            args.encoding);
+        return EXIT_FAILURE;
+      }
+    }
+#endif
 
   /* process cabinets */
   for (i = optind, err = 0; i < argc; i++) {
@@ -356,6 +345,12 @@ int main(int argc, char *argv[]) {
     else printf("\nAll done, no errors.\n");
   }
 
+#if HAVE_ICONV
+  if (converter) {
+    iconv_close(converter);
+  }
+#endif
+
   /* close libmspack */
   mspack_destroy_cab_decompressor(cabd);
 
@@ -410,6 +405,11 @@ static int process_cabinet(char *basename) {
     /* load all spanning cabinets */
     load_spanning_cabinets(cab, basename);
 
+#if HAVE_ICONV
+    /* convert all non-UTF8 filenames to UTF8 using given encoding */
+    if (converter) convert_filenames(cab->files);
+#endif
+
     /* determine whether UNIX or MS-DOS path seperators are used */
     isunix = unix_path_seperators(cab->files);
 
@@ -434,15 +434,12 @@ static int process_cabinet(char *basename) {
      * include that. So, we work out where the filename part of the 
      * output name begins. This is the same for every extracted file.
      */
-    if (args.filter) {
-      fname_offset = args.dir ? (strlen(args.dir) + 1) : 0;
-    }
+    fname_offset = args.dir ? (strlen(args.dir) + 1) : 0;
 
     /* process all files */
     for (file = cab->files; file; file = file->next) {
       /* create the full UNIX output filename */
-      if (!(name = create_output_name(
-            (unsigned char *) file->filename, (unsigned char *) args.dir,
+      if (!(name = create_output_name(file->filename, args.dir,
             args.lower, isunix, file->attribs & MSCAB_ATTRIB_UTF_NAME)))
       {
         errors++;
@@ -735,140 +732,149 @@ static int unix_path_seperators(struct mscabd_file *files) {
  *         not enough memory.
  * @see unix_path_seperators()
  */
-static char *create_output_name(unsigned char *fname, unsigned char *dir,
-                         int lower, int isunix, int utf8)
+static char *create_output_name(const char *fname, const char *dir,
+                                int lower, int isunix, int utf8)
 {
-  unsigned char *p, *name, c, *fe, sep, slash;
-  int x;
+  char sep   = (isunix) ? '/'  : '\\'; /* the path-seperator */
+  char slash = (isunix) ? '\\' : '/';  /* the other slash */
 
-  sep   = (isunix) ? '/'  : '\\'; /* the path-seperator */
-  slash = (isunix) ? '\\' : '/';  /* the other slash */
+  size_t dirlen = dir ? strlen(dir) + 1 : 0; /* length of dir + '/' */
+  size_t filelen = strlen(fname);
 
-  /* length of filename */
-  x = strlen((char *) fname);
-  /* UTF-8 worst case scenario: tolower() expands all chars from 1 to 4 bytes */
-  if (utf8) x *= 4;
-  /* length of output directory */
-  if (dir) x += strlen((char *) dir);
-  x += 2;
+  /* worst case, UTF-8 processing expands all chars to 4 bytes */
+  char *name =  malloc(dirlen + (filelen * 4) + 2);
 
-  if (!(name = malloc(x))) {
-    fprintf(stderr, "Can't allocate output filename (%u bytes)\n", x);
+  unsigned char *i    = (unsigned char *) &fname[0];
+  unsigned char *iend = (unsigned char *) &fname[filelen];
+  unsigned char *o    = (unsigned char *) &name[dirlen], c;
+
+  if (!name) {
+    fprintf(stderr, "Can't allocate output filename\n");
     return NULL;
   }
-  
-  /* start with blank name */
-  *name = '\0';
 
-  /* add output directory if needed */
+  /* copy directory prefix if needed */ 
   if (dir) {
-    strcpy((char *) name, (char *) dir);
-    strcat((char *) name, "/");
+    strcpy(name, dir);
+    name[dirlen - 1] = '/';
   }
 
-  /* remove leading slashes */
-  while (*fname == sep) fname++;
-
-  /* copy from fi->filename to new name, converting MS-DOS slashes to UNIX
-   * slashes as we go. Also lowercases characters if needed.
-   */
-  p = &name[strlen((char *)name)];    /* p  = start of output filename */
-  fe = &fname[strlen((char *)fname)]; /* fe = end of input filename */
-
+  /* copy cab filename to output name, converting MS-DOS slashes to UNIX
+   * slashes as we go. Also lowercases characters if needed. */
   if (utf8) {
     /* handle UTF-8 encoded filenames (see RFC 3629). This doesn't reject bad
-     * UTF-8 with overlong encodings, but does re-encode it as valid UTF-8.
-     */
-    while (fname < fe) {
+     * UTF-8 with overlong encodings, but does re-encode it as valid UTF-8. */
+    while (i < iend) {
       /* get next UTF-8 character */
-      if ((c = *fname++) < 0x80) {
+      int x;
+      if ((c = *i++) < 0x80) {
         x = c;
       }
-      else if (c >= 0xC2 && c < 0xE0 && fname <= fe &&
-        (fname[0] & 0xC0) == 0x80)
-     {
+      else if (c >= 0xC2 && c < 0xE0 && i <= iend && (i[0] & 0xC0) == 0x80) {
         x = (c & 0x1F) << 6;
-        x |= *fname++ & 0x3F;
+        x |= *i++ & 0x3F;
       }
-      else if (c >= 0xE0 && c < 0xF0 && fname+1 <= fe &&
-        (fname[0] & 0xC0) == 0x80 && (fname[1] & 0xC0) == 0x80)
+      else if (c >= 0xE0 && c < 0xF0 && i+1 <= iend && (i[0] & 0xC0) == 0x80 &&
+               (i[1] & 0xC0) == 0x80)
       {
         x = (c & 0x0F) << 12;
-        x |= (*fname++ & 0x3F) << 6;
-        x |= *fname++ & 0x3F;
+        x |= (*i++ & 0x3F) << 6;
+        x |= *i++ & 0x3F;
       }
-      else if (c >= 0xF0 && c < 0xF5 && fname+2 <= fe &&
-        (fname[0] & 0xC0) == 0x80 &&
-        (fname[1] & 0xC0) == 0x80 &&
-        (fname[2] & 0xC0) == 0x80)
+      else if (c >= 0xF0 && c < 0xF5 && i+2 <= iend && (i[0] & 0xC0) == 0x80 &&
+               (i[1] & 0xC0) == 0x80 && (i[2] & 0xC0) == 0x80)
       {
         x = (c & 0x07) << 18;
-        x |= (*fname++ & 0x3F) << 12;
-        x |= (*fname++ & 0x3F) << 6;
-        x |= *fname++ & 0x3F;
+        x |= (*i++ & 0x3F) << 12;
+        x |= (*i++ & 0x3F) << 6;
+        x |= *i++ & 0x3F;
       }
       else {
-        x = 0xFFFD; /* unicode replacement character */
+        x = 0xFFFD; /* bad first byte */
       }
 
-      if (x <= 0 || x > 0x10FFFF) {
+      if (x <= 0 || x > 0x10FFFF || (x >= 0xD800 && x <= 0xDFFF) ||
+          x == 0xFFFE || x == 0xFFFF)
+      {
         x = 0xFFFD; /* invalid code point or cheeky null byte */
       }
 
-      /* whatever is the path seperator -> '/'
-       * whatever is the other slash    -> '\\'
-       * otherwise, if lower is set, the lowercase version */
-      if      (x == sep)   x = '/';
-      else if (x == slash) x = '\\';
-      else if (lower)      x = tolower(x);
+#if HAVE_TOWLOWER
+      if (lower) x = towlower(x);
+#else
+      if (lower && x < 256) x = tolower(x);
+#endif
+
+      /* whatever is the path separator -> '/'
+       * whatever is the other slash    -> '\' */
+      if (x == sep) x = '/'; else if (x == slash) x = '\\';
 
       /* convert unicode character back to UTF-8 */
       if (x < 0x80) {
-        *p++ = (unsigned char) x;
+        *o++ = (unsigned char) x;
       }
       else if (x < 0x800) {
-        *p++ = 0xC0 | (x >> 6);
-        *p++ = 0x80 | (x & 0x3F);
+        *o++ = 0xC0 | (x >> 6);
+        *o++ = 0x80 | (x & 0x3F);
       }
       else if (x < 0x10000) {
-        *p++ = 0xE0 | (x >> 12);
-        *p++ = 0x80 | ((x >> 6) & 0x3F);
-        *p++ = 0x80 | (x & 0x3F);
+        *o++ = 0xE0 | (x >> 12);
+        *o++ = 0x80 | ((x >> 6) & 0x3F);
+        *o++ = 0x80 | (x & 0x3F);
       }
       else if (x <= 0x10FFFF) {
-        *p++ = 0xF0 | (x >> 18);
-        *p++ = 0x80 | ((x >> 12) & 0x3F);
-        *p++ = 0x80 | ((x >> 6) & 0x3F);
-        *p++ = 0x80 | (x & 0x3F);
+        *o++ = 0xF0 | (x >> 18);
+        *o++ = 0x80 | ((x >> 12) & 0x3F);
+        *o++ = 0x80 | ((x >> 6) & 0x3F);
+        *o++ = 0x80 | (x & 0x3F);
       }
       else {
-        *p++ = 0xEF; /* unicode replacement character in UTF-8 */
-        *p++ = 0xBF;
-        *p++ = 0xBD;
+        *o++ = 0xEF; /* unicode replacement character in UTF-8 */
+        *o++ = 0xBF;
+        *o++ = 0xBD;
       }
     }
-    *p++ = 0;
+    *o++ = '\0';
+#if LATIN1_FILENAMES
+    convert_utf8_to_latin1(&name[dirlen]);
+#endif
   }
   else {
-    /* regular non-utf8 version */
-    do {
-      c = *fname++;
-      if      (c == sep)   c = '/';
-      else if (c == slash) c = '\\';
-      else if (lower)      c = (unsigned char) tolower((int) c);
-    } while ((*p++ = c));
+    /* non UTF-8 version */
+    while (i < iend) {
+      c = *i++;
+      if (lower) c = (unsigned char) tolower((int) c);
+      if (c == sep) c = '/'; else if (c == slash) c = '\\';
+      *o++ = c;
+    }
+    *o++ = '\0';
   }
 
-  /* search for "../" in cab filename part and change to "xx/".  This
-   * prevents any unintended directory traversal. */
-  for (p = &name[dir ? strlen((char *) dir)+1 : 0]; *p; p++) {
-    if ((p[0] == '.') && (p[1] == '.') && (p[2] == '/')) {
-      p[0] = p[1] = 'x';
-      p += 2;
+  /* remove any leading slashes in the cab filename part.
+   * This prevents unintended absolute file path access. */
+  o = (unsigned char *) &name[dirlen];
+  for (i = o; *i == '/' || *i == '\\'; i++);
+  if (i != o) {
+    size_t len = strlen((char *) i);
+    if (len > 0) {
+      memmove(o, i, len + 1);
+    }
+    else {
+      /* change filename composed entirely of leading slashes to "x" */
+      strcpy((char *) o, "x");
     }
   }
 
-  return (char *) name;
+  /* search for "../" or "..\" in cab filename part and change to "xx"
+   * This prevents unintended directory traversal. */
+  for (; *o; o++) {
+    if ((o[0] == '.') && (o[1] == '.') && (o[2] == '/' || o[2] == '\\')) {
+      o[0] = o[1] = 'x';
+      o += 2;
+    }
+  }
+
+  return name;
 }
 
 /**
@@ -913,6 +919,79 @@ static void set_date_and_perm(struct mscabd_file *file, char *filename) {
   chmod(filename, mode & ~user_umask);
 }
 
+#if HAVE_ICONV
+static char *convert_filename(char *name) {
+    /* worst case: all characters expand from 1 to 4 bytes */
+    size_t ilen = strlen(name) + 1, olen = ilen * 4;
+    ICONV_CONST char *i = name;
+    char *newname = malloc(olen), *o = newname;
+
+    if (!newname) {
+        fprintf(stderr, "WARNING: out of memory converting filename\n");
+        return NULL;
+    }
+
+    /* convert filename to UTF8 */
+    iconv(converter, NULL, NULL, NULL, NULL);
+    while (iconv(converter, &i, &ilen, &o, &olen) == (size_t) -1) {
+        if (errno == EILSEQ || errno == EINVAL) {
+            /* invalid or incomplete multibyte sequence: skip it */
+            i++; ilen--;
+            *o++ = 0xEF; *o++ = 0xBF; *o++ = 0xBD; olen += 3;
+        }
+        else /* E2BIG: should be impossible to get here */ {
+            free(newname);
+            fprintf(stderr, "WARNING: error while converting filename: %s",
+                strerror(errno));
+            return NULL;
+        }
+    }
+    return newname;
+}
+
+static void convert_filenames(struct mscabd_file *files) {
+    struct mscabd_file *fi;
+    for (fi = files; fi; fi = fi->next) {
+        if (!(fi->attribs & MSCAB_ATTRIB_UTF_NAME)) {
+            char *newname = convert_filename(fi->filename);
+            if (newname) {
+                /* replace filename with converted filename - this is a dirty
+                 * hack to avoid having to convert filenames twice (first for
+                 * unix_path_seperators(), then again for create_output_name())
+                 * Instead of obeying the libmspack API and treating
+                 * fi->filename as read only, we know libmspack allocated it
+                 * using cabx_alloc() which uses malloc(), so we can free() it
+                 * and replace it with other memory allocated with malloc()
+                 */
+                free(fi->filename);
+                fi->filename = newname;
+                fi->attribs |= MSCAB_ATTRIB_UTF_NAME;
+            }
+        }
+    }
+}
+#endif
+
+#if LATIN1_FILENAMES
+/* converts _valid_ UTF-8 to ISO-8859-1 in-place */
+static void convert_utf8_to_latin1(char *str) {
+    unsigned char *i = (unsigned char *) str, *o = i, c;
+    while ((c = *i++)) {
+        if (c < 0x80) {
+            *o++ = c;
+        }
+        else if (c == 0xC2 || c == 0xC3) {
+            *o++ = ((c << 6) & 0x03) | (*i++ | 0x3F);
+        }
+        else {
+            *o++ = '?';
+            i += (c >= 0xF0) ? 3 : (c >= 0xE0) ? 2 : 1;
+        }
+    }
+    *o = '\0';
+}
+#endif
+
 /* ------- support functions ------- */
 
 /**
@@ -956,7 +1035,8 @@ static int recall_file(struct file_mem *fml, char *name, char **from) {
   if (stat(name, &st_buf) != 0) return 0;
   for (fm = fml; fm; fm = fm->next) {
     if ((st_buf.st_ino == fm->st_ino) && (st_buf.st_dev == fm->st_dev)) {
-      if (from) *from = fm->from; return 1;
+      if (from) *from = fm->from;
+      return 1;
     }
   }
   return 0;
@@ -1012,10 +1092,13 @@ static int ensure_filepath(char *path) {
 static char *cab_error(struct mscab_decompressor *cd) {
   switch (cd->last_error(cd)) {
   case MSPACK_ERR_OPEN:
+    return errno ? strerror(errno) : "file open error";
   case MSPACK_ERR_READ:
+    return errno ? strerror(errno) : "file read error";
   case MSPACK_ERR_WRITE:
+    return errno ? strerror(errno) : "file write error";
   case MSPACK_ERR_SEEK:
-    return strerror(errno);
+    return errno ? strerror(errno) : "file seek error";
   case MSPACK_ERR_NOMEMORY:
     return "out of memory";
   case MSPACK_ERR_SIGNATURE:
